from abc import ABC, abstractmethod
import numpy as np
from typing import List
from cryolo.utils import BoundBox
from dataclasses import dataclass
import os

@dataclass
class ImageBoxData:
    indentifer : str
    group : str
    data : np.array

class BoxDataProvider(ABC):

    @abstractmethod
    def get_boxes(self) -> List[ImageBoxData]:
        """
        :return: List of box coordinates as numpy array.
        """
        pass

    @staticmethod
    def path_matching_filename_eq(patha: str, pathb: str) -> bool:
        # FILL THAT TOMORROW
        filenamea = os.path.splitext(os.path.basename(patha))[0]
        filenameb = os.path.splitext(os.path.basename(pathb))[0]
        return filenamea == filenameb

    @staticmethod
    def group_method_lastfolder(path: str) -> str:
        return os.path.basename(os.path.dirname(path))

    @staticmethod
    def boxes_to_nparray(boxes : List[BoundBox]) -> np.array:
        """
        :param boxes: List of Bounding boxes
        :return: np aarray with columns x,y,z,width,height,depth for 3D and x,y width,height for 2D
        """

        is2D = np.any([b.z==None for b in boxes])

        if is2D:
            data = np.zeros(shape=(len(boxes), 4))
        else:
            data = np.zeros(shape=(len(boxes), 6))

        for boxi, box in enumerate(boxes):
            data[boxi, 0] = box.x
            data[boxi, 1] = box.y
            if is2D:
                data[boxi, 2] = box.w
                data[boxi, 3] = box.h
            else:
                data[boxi, 2] = box.z
                data[boxi, 3] = box.w
                data[boxi, 4] = box.h
                data[boxi, 5] = box.depth

        return data


#
# COPYRIGHT
#
# All contributions by Ngoc Anh Huyn:
# Copyright (c) 2017, Ngoc Anh Huyn.
# All rights reserved.
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#

import multiprocessing
import os
import copy

from PIL import Image
import keras
import numpy as np
import tensorflow as tf
from keras.callbacks import EarlyStopping, TensorBoard
from keras.layers import Reshape, Conv2D, Input, Lambda, Dropout
from keras.models import Model
from keras.optimizers import Adam

import cryolo.utils as utils
import cryolo
from cryolo.MultiGPUModelCheckpoint import MultiGPUModelCheckpoint
from cryolo.ExtendedModelCheckpoint import ExtendedModelCheckpoint
from cryolo.warmup_callback import WarmupCallback
from cryolo.SGDRSchedular import SGDRScheduler
from cryolo.backend import FullYoloFeature, CrYoloFeature, PhosaurusNetFeature, BaseFeatureExtractor
from cryolo.patchwisebatchgenerator2 import PatchwiseBatchGenerator
from cryolo.utils import BoundBox
from typing import Sequence
warmup_bs = keras.backend.variable(1.0)


class FeatureExtractorFactory():

    @staticmethod
    def create_feature_extractor(architecture : str , input_size : Sequence, input_depth : int, backend_weights : str) -> BaseFeatureExtractor:
        try:
            if architecture == "YOLO":
                return FullYoloFeature(
                    input_size, input_depth, backend_weights
                )
            elif architecture == "crYOLO":
                return CrYoloFeature(
                    input_size, input_depth, backend_weights
                )
            elif architecture == "PhosaurusNet":
                return PhosaurusNetFeature(
                    input_size, input_depth, backend_weights
                )
            else:
                raise Exception(
                    "Architecture not supported! Only support YOLO and crYOLO at the moment!"
                )
        except tf.errors.ResourceExhaustedError as e:
            print(e)
            import sys

            sys.exit(
                "Seems that crYOLO was not able to allocate memory. "
                "Is your GPU currenlty used? Check it with nivdia-smi"
            )
        return None


class YOLO:
    def __init__(
        self,
        architecture,
        input_size,
        max_box_per_image,
        anchors,
        backend_weights=None,
        uniitest=False,
        experimental_loss=False,
        pretrained_weights=None,
        pretrained_weights_init=False, #only use for pre initialization during training
        fine_tune=False,
        num_fine_tune_layers=2,
        input_depth=1,
        labels=["particle"],
        filament_model=False
    ):
        self.input_size = input_size
        self.input_depth = input_depth
        self.labels = list(labels)
        self.nb_class = len(self.labels)
        self.nb_box = int(len(anchors) / 2)
        self.class_wt = np.ones(self.nb_class, dtype="float32")
        self.anchors = anchors
        self.max_box_per_image = max_box_per_image
        self.nms_threshold = None
        self.obj_threshold = None
        self.experimental_loss = experimental_loss
        self.features = None
        self.fine_tune = fine_tune
        self.num_fine_tune_layers = num_fine_tune_layers


        ##########################
        # Make the model
        ##########################

        # make the feature extractor layers
        if not uniitest:
            if type(self.input_size) is list:
                self.input_size = self.input_size[::-1]
            else:
                self.input_size = [self.input_size,self.input_size]
            input_image = Input(
                shape=(self.input_size[1], self.input_size[0], self.input_depth),
                dtype="float32",
            )
            self.is_filament_model = filament_model
            self.true_boxes = Input(shape=(1, 1, 1, max_box_per_image, 4 + self.is_filament_model))

            self.feature_extractor = FeatureExtractorFactory.create_feature_extractor(architecture=architecture,
                                                                                      input_size=self.input_size,
                                                                                      input_depth=self.input_depth,
                                                                                      backend_weights=backend_weights)

            self.grid_rows, self.grid_columns = self.feature_extractor.get_output_shape()

            features = self.feature_extractor.extract(input_image)

            # make the object detection layer
            self.features = Dropout(0.2, name="Dropout1")(features)
            postfix = ""
            if self.is_filament_model:
                postfix = "_filament"
            output = Conv2D(
                self.nb_box * (4 + 1 + self.nb_class + self.is_filament_model),
                (1, 1),
                strides=(1, 1),
                padding="same",
                name="conv_23" + postfix,
                kernel_initializer="lecun_normal",
            )(features)
            output = Reshape(
                (self.grid_rows, self.grid_columns, self.nb_box, 4 + 1 + self.nb_class + self.is_filament_model)
            )(output)
            output = Lambda(lambda args: args[0])([output, self.true_boxes])
            self.model = Model([input_image, self.true_boxes], output)

            # initialize the weights of the detection layer
            layer = self.model.layers[-4]
            weights = layer.get_weights()

            new_kernel = np.random.normal(size=weights[0].shape) / (
                    self.grid_rows * self.grid_columns
            )
            new_bias = np.random.normal(size=weights[1].shape) / (
                    self.grid_rows * self.grid_columns
            )

            layer.set_weights([new_kernel, new_bias])

            if pretrained_weights is not None and os.path.exists(pretrained_weights):
                print("\n Load weights: ", pretrained_weights)
                self.load_weights(
                    weight_path=pretrained_weights,
                    do_fine_tune=self.fine_tune,
                    num_layers=self.num_fine_tune_layers,
                    is_init=pretrained_weights_init
                )

    def custom_loss(self, y_true, y_pred):
        """
        For y_true: The last dimension saves 6 values:
        [0]: x-position Relative to cell width. E.g.: if the cell with is 32, then the value are from 0 - 32
        [1]: y-position
        [2]: box-width In units of cell width, relative to downscaled image. E.g.: if the cell width is 25, and the
        estimataed box size 200/4=50, then it would be 2
        [3]: box-height analog to box-width
        [4]: confidence. contains 1 and 0 for y_true and -Inf to 1 for y_pred
        [5]: class-probabilites
        :param y_true: Tensor with shape [batch_size, grid_width, grid_height, num_anchor_boxes, 6]
        :param y_pred:
        :return:
        """

        mask_shape = tf.shape(y_true)[:4]

        # Generates a grid (e.g. 32 x 32) with coordinates
        cell_x = tf.to_float(
            tf.reshape(
                tf.tile(tf.range(self.grid_columns), [self.grid_rows]),
                (1, self.grid_rows, self.grid_columns, 1, 1),
            )
        )
        cell_y = tf.to_float(
            tf.reshape(tf.tile(tf.range(self.grid_rows), [self.grid_columns]), (1, self.grid_columns, self.grid_rows, 1, 1)))
        cell_y = tf.transpose(cell_y, (0, 2, 1, 3, 4))

        cell_grid = tf.tile(
            tf.concat([cell_x, cell_y], -1), [self.batch_size, 1, 1, self.nb_box, 1]
        )

        coord_mask = tf.zeros(mask_shape)
        conf_mask = tf.zeros(mask_shape)

        seen = tf.Variable(0.0)
        total_recall = tf.Variable(0.0)

        #
        # Adjust prediction
        #

        # adjust x and y
        pred_box_xy = tf.sigmoid(y_pred[..., :2]) + cell_grid #coordinate corner + relative box position

        # adjust w and h
        pred_box_wh = tf.exp(y_pred[..., 2:4]) * np.reshape(
            self.anchors, [1, 1, 1, self.nb_box, 2]
        )



        # adjust confidence
        pred_box_conf = tf.sigmoid(y_pred[..., 4])

        if self.is_filament_model:
            pred_box_angle = tf.sigmoid(y_pred[..., -1]) # from 0 to 1 -> 0 to PI
            pred_box_class = y_pred[..., 5:-1]
        else:
            # adjust class probabilities
            pred_box_class = y_pred[..., 5:]

        #
        # Adjust ground truth
        #

        # adjust x and y
        true_box_xy = y_true[..., 0:2]  # relative position to the containing cell

        # adjust w and h
        true_box_wh = y_true[
            ..., 2:4
        ]  # number of cells across, horizontally and vertically

        # adjust confidence
        true_wh_half = true_box_wh / 2.0
        true_mins = true_box_xy - true_wh_half
        true_maxes = true_box_xy + true_wh_half

        pred_wh_half = pred_box_wh / 2.0
        pred_mins = pred_box_xy - pred_wh_half
        pred_maxes = pred_box_xy + pred_wh_half

        intersect_mins = tf.maximum(pred_mins, true_mins)
        intersect_maxes = tf.minimum(pred_maxes, true_maxes)
        intersect_wh = tf.maximum(intersect_maxes - intersect_mins, 0.0)
        intersect_areas = intersect_wh[..., 0] * intersect_wh[..., 1]

        true_areas = true_box_wh[..., 0] * true_box_wh[..., 1]
        pred_areas = pred_box_wh[..., 0] * pred_box_wh[..., 1]

        union_areas = pred_areas + true_areas - intersect_areas
        iou_scores = tf.truediv(intersect_areas, union_areas)

        true_box_conf = iou_scores * y_true[..., 4]  # Page 2 in YOLO Paper

        # adjust class probabilities
        if self.is_filament_model:
            true_box_class = tf.argmax(
                y_true[..., 5:-1], -1
            )
            true_box_angle = y_true[..., -1]
        else:
            true_box_class = tf.argmax(
                y_true[..., 5:], -1
            )  # is always zero if there is only one class

        #
        # Determine the masks
        #

        # coordinate mask: simply the position of the ground truth boxes (the predictors)
        coord_mask = tf.expand_dims(y_true[..., 4], axis=-1) * self.coord_scale

        # confidence mask: penelize predictors + penalize boxes with low IOU
        # penalize the confidence of the boxes, which have IOU with some ground truth box < 0.6
        true_xy = self.true_boxes[..., 0:2]
        true_wh = self.true_boxes[..., 2:4]

        true_wh_half = true_wh / 2.0
        true_mins = true_xy - true_wh_half
        true_maxes = true_xy + true_wh_half

        pred_xy = tf.expand_dims(pred_box_xy, 4)
        pred_wh = tf.expand_dims(pred_box_wh, 4)

        pred_wh_half = pred_wh / 2.0
        pred_mins = pred_xy - pred_wh_half
        pred_maxes = pred_xy + pred_wh_half

        intersect_mins = tf.maximum(pred_mins, true_mins)
        intersect_maxes = tf.minimum(pred_maxes, true_maxes)
        intersect_wh = tf.maximum(intersect_maxes - intersect_mins, 0.0)
        intersect_areas = intersect_wh[..., 0] * intersect_wh[..., 1]

        true_areas = true_wh[..., 0] * true_wh[..., 1]
        pred_areas = pred_wh[..., 0] * pred_wh[..., 1]

        union_areas = pred_areas + true_areas - intersect_areas
        iou_scores = tf.truediv(intersect_areas, union_areas)

        best_ious = tf.reduce_max(iou_scores, axis=4)

        # penalize the confidence of the boxes, which are reponsible for corresponding ground truth box
        if self.experimental_loss:
            conf_mask_noobj = (
                conf_mask
                + tf.to_float(best_ious < 0.6)
                * (1 - y_true[..., 4])
                * self.no_object_scale
            )
            conf_mask_obj = conf_mask + y_true[..., 4] * self.object_scale
        else:
            conf_mask = (
                conf_mask
                + tf.to_float(best_ious < 0.6)
                * (1 - y_true[..., 4])
                * self.no_object_scale
            )
            conf_mask = conf_mask + y_true[..., 4] * self.object_scale

        # class mask: simply the position of the ground truth boxes (the predictors)
        class_mask = (
            y_true[..., 4] * tf.gather(self.class_wt, true_box_class) * self.class_scale
        )

        #
        # Warm-up training
        #
        no_boxes_mask = tf.to_float(coord_mask < self.coord_scale / 2.0)
        seen = tf.assign_add(seen, 1.0)

        true_box_xy, true_box_wh, coord_mask = tf.cond(
            tf.less(seen, warmup_bs),
            lambda: [
                true_box_xy + (0.5 + cell_grid) * no_boxes_mask,
                true_box_wh
                + tf.ones_like(true_box_wh)
                * np.reshape(self.anchors, [1, 1, 1, self.nb_box, 2])
                * no_boxes_mask,
                tf.ones_like(coord_mask),
            ],
            lambda: [true_box_xy, true_box_wh, coord_mask],
        )

        #
        # Finalize the loss
        #

        nb_coord_box = tf.reduce_sum(tf.to_float(coord_mask > 0.0))

        if self.experimental_loss:
            nb_conf_box_obj = tf.reduce_sum(tf.to_float(conf_mask_obj > 0.0))
            nb_conf_box_noobj = tf.reduce_sum(tf.to_float(conf_mask_noobj > 0.0))
        else:
            nb_conf_box = tf.reduce_sum(tf.to_float(conf_mask > 0.0))

        nb_class_box = tf.reduce_sum(tf.to_float(class_mask > 0.0))

        loss_xy = (
            tf.reduce_sum(tf.square(true_box_xy - pred_box_xy) * coord_mask)
            / (nb_coord_box + 1e-6)
            / 2.0
        )
        loss_wh = (
            tf.reduce_sum(tf.square(true_box_wh - pred_box_wh) * coord_mask)
            / (nb_coord_box + 1e-6)
            / 2.0
        )

        if self.is_filament_model:
            # I should do it similar to confidence, but not sure why the
            # division by 2 happens
            #loss_angle = true_box_angle - pred_box_angle
            '''
            diff_angles_a = tf.abs(true_box_angle - pred_box_angle)
            # Reason for +-0.5: The 0.5 is actuall +- PI, but the true angles are in the range from
            # 0 to 1 instead of 0 to 2PI because they are normalized. 0.5 is also actually +- PI
            diff_angles_b = tf.abs(true_box_angle - pred_box_angle + 1)
            diff_angles_c = tf.abs(true_box_angle - pred_box_angle - 1)

            diff_angles = tf.minimum(tf.math.minimum(diff_angles_a,diff_angles_b),diff_angles_c)
            '''
            dot = tf.square(
                tf.cos(true_box_angle*np.pi)*tf.cos(pred_box_angle*np.pi)+
                tf.sin(true_box_angle*np.pi)*tf.sin(pred_box_angle*np.pi)
            )
            loss_angle = (
                tf.reduce_sum((1-dot) * coord_mask[:,:,:,0,:]) # conf_mask
                / (nb_coord_box + 1e-6) # nb_conf_box
                / 2
            )


        if self.experimental_loss:
            loss_conf_obj = (
                tf.reduce_sum(tf.square(true_box_conf - pred_box_conf) * conf_mask_obj)
                / (nb_conf_box_obj + 1e-6)
                / 2.0
            )
            loss_conf_noobj = (
                tf.reduce_sum(
                    tf.square(true_box_conf - pred_box_conf) * conf_mask_noobj
                )
                / (nb_conf_box_noobj + 1e-6)
                / 2.0
            )
        else:
            loss_conf = (
                tf.reduce_sum(tf.square(true_box_conf - pred_box_conf) * conf_mask)
                / (nb_conf_box + 1e-6)
                / 2.0
            )
        loss_class = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels=true_box_class, logits=pred_box_class
        )
        loss_class = tf.reduce_sum(loss_class * class_mask) / (nb_class_box + 1e-6)

        if self.experimental_loss:
            loss = loss_xy + loss_conf_obj + loss_conf_noobj + loss_class + loss_wh
        else:
            if self.is_filament_model:
                loss = loss_xy + loss_conf + loss_class + loss_wh + loss_angle
            else:
                loss = loss_xy + loss_conf + loss_class + loss_wh

        if self.debug:
            nb_true_box = tf.reduce_sum(y_true[..., 4])
            nb_pred_box = tf.reduce_sum(
                tf.to_float(true_box_conf > 0.5) * tf.to_float(pred_box_conf > 0.3)
            )

            current_recall = nb_pred_box / (nb_true_box + 1e-6)
            total_recall = tf.assign_add(total_recall, current_recall)
            try:
                # New tensorflow > 1.10.1
                if self.experimental_loss:
                    print_op = tf.print(
                        "\n",
                        "\tLoss XY \t",
                        loss_xy,
                        "\n",
                        "\tLoss WH \t",
                        loss_wh,
                        "\n",
                        "\tLoss Obj \t",
                        loss_conf_obj,
                        "\n",
                        "\tLoss NoObj \t",
                        loss_conf_noobj,
                        "\n",
                        "\tTotal Loss \t",
                        loss,
                        "\n",
                        "\tCurrent Recall \t",
                        current_recall,
                        "\n",
                        "\tAverage Recall \t",
                        total_recall / seen,
                        "\n",
                    )
                else:
                    print_op = tf.print(
                        "\n",
                        "\tLoss XY \t",
                        loss_xy,
                        "\n",
                        "\tLoss WH \t",
                        loss_wh,
                        "\n",
                        "\tLoss Conf \t",
                        loss_conf,
                        "\n",
                        "\tLoss Angle \t",
                         loss_angle,
                        "\n",
                        "\tTotal Loss \t",
                        loss,
                        "\n",
                        "\tCurrent Recall \t",
                        current_recall,
                        "\n",
                        "\tAverage Recall \t",
                        total_recall / seen,
                        "\n",
                    )
                with tf.control_dependencies([print_op]):
                    loss = tf.identity(loss)
            except:
                # Old tensorflow == 1.10.1
                loss = tf.Print(loss, [loss_xy], message="\nLoss XY \t", summarize=1000)
                loss = tf.Print(loss, [loss_wh], message="Loss WH \t", summarize=1000)
                if self.experimental_loss:
                    loss = tf.Print(
                        loss, [loss_conf_obj], message="Loss Obj \t", summarize=1000
                    )
                    loss = tf.Print(
                        loss, [loss_conf_noobj], message="Loss NoObj \t", summarize=1000
                    )
                else:
                    loss = tf.Print(
                        loss, [loss_conf], message="Loss Conf \t", summarize=1000
                    )
                # loss = tf.Print(loss, [loss_class], message='Loss Class \t', summarize=1000)
                loss = tf.Print(loss, [loss], message="Total Loss \t", summarize=1000)
                loss = tf.Print(
                    loss, [current_recall], message="Current Recall \t", summarize=1000
                )
                loss = tf.Print(
                    loss,
                    [total_recall / seen],
                    message="Average Recall \t",
                    summarize=1000,
                )

        return loss

    def load_weights(self, weight_path: str, do_fine_tune: bool=False, num_layers: int=2, is_init: bool=False):
        '''

        :param weight_path: Path to weights
        :param do_fine_tune: Boolean flag if only fine tuning is done
        :param num_layers: Number of layers to fine tune
        :param is_init: Boolean flag. True if the weights are only loaded for pre initialization during training.
        '''

        try:
            if is_init:
                self.model.load_weights(weight_path, by_name=True)
            else:
                self.model.load_weights(weight_path)

            ###############################
            # If wanted, only train the last two layers of the network.
            ###############################
            if do_fine_tune:
                print("Do fine tuning!")
                for layer_index, layer in enumerate(self.model.layers):
                    if layer.name != "conv_23" and "model_" not in layer.name:
                        layer.trainable = False

                #
                # Find number of convolutional layers
                #
                num_conv_layers = 0

                for layer_index, layer in enumerate(
                    self.model.get_layer(index=1).layers
                ):
                    if "conv_" in layer.name:
                        num_conv_layers = num_conv_layers + 1

                #
                # Define trainable layers
                #
                free_layers = []
                for i in range(num_layers - 1):
                    layer_index = num_conv_layers - i
                    free_layers.append("conv_" + str(layer_index))
                    free_layers.append("conv_" + str(layer_index) + "_")

                #
                # Freeze other layers
                #
                for layer_index, layer in enumerate(
                    self.model.get_layer(index=1).layers
                ):
                    if layer.name not in free_layers:
                        layer.trainable = False

        except ValueError:
            ###############################
            # This is a workaround for a keras bug loading weights from nested models with fixed layers
            ###############################
            for layer_index, layer in enumerate(self.model.layers):
                if layer.name != "conv_23" and "model_" not in layer.name:
                    layer.trainable = False

            #
            # Find number of convolutional layers
            #
            num_conv_layers = 0

            for layer_index, layer in enumerate(self.model.get_layer(index=1).layers):
                if "conv_" in layer.name:
                    num_conv_layers = num_conv_layers + 1

            #
            # Define trainable layers
            #
            free_layers = []
            for i in range(num_layers - 1):
                layer_index = num_conv_layers - i
                free_layers.append("conv_" + str(layer_index))
                free_layers.append("conv_" + str(layer_index) + "_")

            #
            # Freeze other layers
            #
            for layer_index, layer in enumerate(self.model.get_layer(index=1).layers):
                if layer.name not in free_layers:
                    layer.trainable = False

                else:
                    print("Make trainable:", layer.name)

            self.model.load_weights(weight_path)

    def predict(
        self,
        image_tiles,
        tiles_coord,
        image_dimensions,
        obj_threshold=0.3,
        nms_threshold=0.3,
        num_patches=1,
        normalize_margin=0,
        normalization="STANDARD",
    ):

        self.nms_threshold = nms_threshold
        self.obj_threshold = obj_threshold
        resized_normalized_array = None
        norm_func = utils.get_normalization_func(normalization)
        for i in range(len(tiles_coord)):
            img_tmp = image_tiles[i, :, :]

            img_tmp = np.array(
                Image.fromarray(img_tmp).resize(
                    (self.input_size[0], self.input_size[1]), resample=Image.BILINEAR
                )
            )
            # cv2.resize(img_tmp, (self.input_size, self.input_size))

            if resized_normalized_array is None:
                resized_normalized_array = np.empty(
                    shape=(image_tiles.shape[0], img_tmp.shape[0], img_tmp.shape[1])
                )

            resized_normalized_array[i, :, :] = norm_func(
                img_tmp, margin_size=normalize_margin
            )

        input_image = resized_normalized_array
        input_image = np.expand_dims(input_image, 3)
        dummy_array = np.zeros(
            (input_image.shape[0], 1, 1, 1, self.max_box_per_image, 4+self.is_filament_model)
        )
        netout = self.model.predict([input_image, dummy_array])
        boxes_nms, boxes_unfiltered = self.decode_netout_fast(
            netout,
            image_tiles,
            tiles_coord,
            image_dimensions,
            obj_threshold,
            num_patches,
        )

        return boxes_nms, boxes_unfiltered

    def decode_netout_fast(
        self,
        netout_batch,
        image_tiles,
        tiles_coord,
        image_dim,
        obj_threshold=0.3,
        num_patches=1,
    ):
        nr_patches = num_patches * num_patches
        boxes_per_image = []
        boxes = []
        sigm = self.sigmoid
        for netout_index in range(netout_batch.shape[0]):
            tile_shape = image_tiles[netout_index].shape
            netout = netout_batch[netout_index, :, :, :, :]
            grid_h, grid_w, nb_box = netout.shape[:3]
            # decode the output by the network
            netout[..., 4] = sigm(netout[..., 4]) # Confidence
            if self.is_filament_model:
                netout[..., 5:-1] = netout[..., 4][..., np.newaxis] * self.softmax(
                    netout[..., 5:-1]
                )

                netout[..., 5:-1] *= netout[..., 5:-1] > min(0.1, obj_threshold)
            else:
                netout[..., 5:] = netout[..., 4][..., np.newaxis] * self.softmax(
                    netout[..., 5:]
                )

                netout[..., 5:] *= netout[..., 5:] > min(0.1, obj_threshold)

            # Transform data
            range_grid_h = np.arange(grid_h)
            range_grid_w = np.arange(grid_w)
            range_grid_w_np = np.array(range_grid_w)
            range_grid_h_np = np.array(range_grid_h)

            netout[:, :, 0, 0] = (range_grid_w_np + sigm(netout[:, :, 0, 0])) / grid_w

            netout[:, :, 0, 0] = (
                netout[:, :, 0, 0] * tile_shape[1]
                + float(tiles_coord[netout_index][0].start)
            ) / image_dim[1]

            netout[:, :, 0, 1] = (
                range_grid_h_np.reshape(range_grid_h_np.shape[0], 1)
                + sigm(netout[:, :, 0, 1])
            ) / grid_h

            netout[:, :, 0, 1] = (
                netout[:, :, 0, 1] * tile_shape[0]
                + float(tiles_coord[netout_index][1].start)
            ) / image_dim[0]

            netout[:, :, 0, 2] = (
                self.anchors[0] * np.exp(netout[:, :, 0, 2]) / grid_w
            )  # unit: image width
            netout[:, :, 0, 2] = netout[:, :, 0, 2] * tile_shape[1] / image_dim[1]

            netout[:, :, 0, 3] = (
                self.anchors[1] * np.exp(netout[:, :, 0, 3]) / grid_h
            )  # unit: image height
            netout[:, :, 0, 3] = netout[:, :, 0, 3] * tile_shape[0] / image_dim[0]

            if self.is_filament_model:
                netout[..., 6] = sigm(netout[..., 6])*np.pi

            # Create boxes data
            for row in range(grid_h):
                for col in range(grid_w):
                    for b in range(nb_box):
                        if self.is_filament_model:
                            if np.sum(netout[row, col, b, 5:-1]) > 0:
                                bb = BoundBox(
                                    x=netout[row, col, b, 0],
                                    y=netout[row, col, b, 1],
                                    w=netout[row, col, b, 2],
                                    h=netout[row, col, b, 3],
                                    c=netout[row, col, b, 4],
                                    classes=netout[row, col, b, 5:-1],
                                )
                                bb.meta={"angle":netout[row, col, b, 6]}
                                #print("Angle:", netout[row, col, b, 6])
                                boxes.append(bb)
                        else:
                            if np.sum(netout[row, col, b, 5:]) > 0:
                                bb = BoundBox(
                                        x=netout[row, col, b, 0],
                                        y=netout[row, col, b, 1],
                                        w=netout[row, col, b, 2],
                                        h=netout[row, col, b, 3],
                                        c=netout[row, col, b, 4],
                                        classes=netout[row, col, b, 5:],
                                    )
                                boxes.append(bb)
            '''
            boxes.extend(
                [
                    BoundBox(
                        x=netout[row, col, b, 0],
                        y=netout[row, col, b, 1],
                        w=netout[row, col, b, 2],
                        h=netout[row, col, b, 3],
                        c=netout[row, col, b, 4],
                        classes=netout[row, col, b, 5:],
                    )
                    for row in range(grid_h)
                    for col in range(grid_w)
                    for b in range(nb_box)
                    if np.sum(netout[row, col, b, 5:]) > 0
                ]
            )
            '''
            if (netout_index + 1) % nr_patches == 0:
                boxes_per_image.append(boxes)

                boxes = []

        boxes_nms = [
            utils.non_maxima_suppress_fast(
                boxes, self.nb_class, self.nms_threshold, min(0.1, obj_threshold)
            )
            for boxes in boxes_per_image
        ]

        boxes_thresholded = [
            [copy.deepcopy(box) for box in boxes if box.c > obj_threshold]
            for boxes in boxes_nms
        ]

        """
        for boxes in boxes_less_filtered:
            filt_box = [[box for box in boxes if box.c > obj_threshold] for boxes in boxes_less_filtered]
            boxes_nms.append(filt_box)
        """
        return boxes_thresholded, boxes_nms

    def decode_netout(
        self,
        netout_batch,
        image_tiles,
        tiles_coord,
        image_dim,
        obj_threshold=0.3,
        num_patches=1,
    ):
        nr_patches = num_patches * num_patches
        boxes_per_image = []
        boxes = []
        sigm = self.sigmoid
        for netout_index in range(netout_batch.shape[0]):

            netout = netout_batch[netout_index, :, :, :, :]
            grid_h, grid_w, nb_box = netout.shape[:3]

            # decode the output by the network
            netout[..., 4] = sigm(netout[..., 4])
            netout[..., 5:] = netout[..., 4][..., np.newaxis] * self.softmax(
                netout[..., 5:]
            )

            netout[..., 5:] *= netout[..., 5:] > min(0.01, obj_threshold)

            for row in range(grid_h):
                for col in range(grid_w):
                    for b in range(nb_box):
                        # from 4th element onwards are confidence and class classes
                        classes = netout[row, col, b, 5:]

                        if np.sum(classes) > 0:
                            # first 4 elements are x, y, w, and h
                            box_x, box_y, box_width, box_height = netout[
                                row, col, b, :4
                            ]

                            box_x = (
                                col + sigm(box_x)
                            ) / grid_w  # center position, unit: image width
                            box_y = (
                                row + sigm(box_y)
                            ) / grid_h  # center position, unit: image height

                            box_width = (
                                self.anchors[2 * b + 0] * np.exp(box_width) / grid_w
                            )  # unit: image width
                            box_height = (
                                self.anchors[2 * b + 1] * np.exp(box_height) / grid_h
                            )  # unit: image height
                            confidence = netout[row, col, b, 4]

                            box = BoundBox(
                                box_x, box_y, box_width, box_height, confidence, classes
                            )

                            tile_shape = image_tiles[netout_index].shape
                            box.x = (
                                box.x * tile_shape[1]
                                + float(tiles_coord[netout_index][0].start)
                            ) / image_dim[1]
                            box.y = (
                                box.y * tile_shape[0]
                                + float(tiles_coord[netout_index][1].start)
                            ) / image_dim[0]

                            box.w = box.w * tile_shape[1] / image_dim[1]
                            box.h = box.h * tile_shape[0] / image_dim[0]
                            boxes.append(box)
            if (netout_index + 1) % nr_patches == 0:
                boxes_per_image.append(boxes)
                boxes = []

        boxes_nms = [
            utils.non_maxima_suppress_fast(
                boxes, self.nb_class, self.nms_threshold, self.obj_threshold
            )
            for boxes in boxes_per_image
        ]

        boxes_less_filtered = [
            utils.non_maxima_suppress_fast(
                boxes, self.nb_class, self.nms_threshold, min(0.01, obj_threshold)
            )
            for boxes in boxes_per_image
        ]
        return boxes_nms, boxes_less_filtered

    def sigmoid(self, x):
        return 1.0 / (1.0 + np.exp(-x))

    def softmax(self, x, axis=-1, t=-100.0):
        x = x - np.max(x)

        if np.min(x) < t:
            x = x / np.min(x) * t

        e_x = np.exp(x)

        return e_x / e_x.sum(axis, keepdims=True)

    def train(
        self,
        train_imgs,  # the list of images to train the model
        valid_imgs,  # the list of images used to validate the model
        train_times,  # the number of time to repeat the training set, often used for small datasets
        valid_times,  # the number of times to repeat the validation set, often used for small datasets
        nb_epoch,  # number of epoches
        learning_rate,  # the learning rate
        batch_size,  # the size of the batch
        warmup_epochs,  # number of initial batches to let the model familiarize with the new dataset
        object_scale,
        no_object_scale,
        coord_scale,
        class_scale,
        cryolo_mode,
        parallel_model=None,
        saved_weights_name="best_weights.h5",
        debug=False,
        log_path="~/logs/",
        early_stop_thresh=3,
        num_patches=1,
        warm_restarts=False,
        overlap_patches=0,
        num_cpus=-1,
        do_augmentation=True,
        do_multiprocessing=True,
        normalization="STANDARD",
        filament_model=False
    ):

        self.batch_size = batch_size
        self.warmup_epochs = warmup_epochs
        global warmup_bs
        keras.backend.set_value(
            warmup_bs,
            warmup_epochs
            * (
                train_times
                * (len(train_imgs) * num_patches * num_patches / batch_size + 1)
                + valid_times
                * (len(valid_imgs) * num_patches * num_patches / batch_size + 1)
            ),
        )
        """  
        
        """
        self.object_scale = object_scale
        self.no_object_scale = no_object_scale
        self.coord_scale = coord_scale
        self.class_scale = class_scale

        self.debug = debug
        self.do_multiprocessing = do_multiprocessing

        # Create save weights path
        saved_weights_filename, saved_weights_extension = os.path.splitext(
            saved_weights_name
        )
        saved_weights_name = saved_weights_filename + "_tmp" + saved_weights_extension
        try:
            os.makedirs(os.path.dirname(saved_weights_name))
        except:
            pass

        # if warmup_epochs > 0:
        #    nb_epoch = (
        #        warmup_epochs
        #    )  # if it's warmup stage, don't train more than warmup_epochs

        ############################################
        # Compile the model
        ############################################

        optimizer = Adam(
            lr=learning_rate, beta_1=0.9, beta_2=0.999, epsilon=1e-08, decay=0.0
        )
        # optimizer = Nadam(lr=learning_rate, beta_1=0.9, beta_2=0.999, epsilon=1e-08)
        # optimizer = SGD(lr=learning_rate, momentum=0.8, decay=0.0, nesterov=True)
        # parallel_model = multi_gpu_model(self.model,gpus=2)
        if parallel_model:
            parallel_model.compile(loss=self.custom_loss, optimizer=optimizer)
        else:
            self.model.compile(loss=self.custom_loss, optimizer=optimizer)

        ############################################
        # Make train and validation generators
        ############################################

        generator_config = {
            "IMAGE_H": self.input_size[1],
            "IMAGE_W": self.input_size[0],
            "GRID_H": self.grid_rows,
            "GRID_W": self.grid_columns,
            "BOX": self.nb_box,
            "LABELS": self.labels,
            "CLASS": len(self.labels),
            "ANCHORS": self.anchors,
            "BATCH_SIZE": self.batch_size,
            "TRUE_BOX_BUFFER": self.max_box_per_image,
            "NUM_PATCHES": num_patches,
        }

        norm_func = utils.get_normalization_func(normalization)

        patch_assignements = np.ones(shape=(num_patches, num_patches))

        train_batch = PatchwiseBatchGenerator(
            train_imgs,
            generator_config,
            patch_assignements,
            cryolo_mode=cryolo_mode,
            norm=norm_func,
            jitter=do_augmentation,
            overlap=overlap_patches,
            name="train",
            train_times=train_times,
            is_filament_training=filament_model
        )

        valid_batch = PatchwiseBatchGenerator(
            valid_imgs,
            generator_config,
            patch_assignements,
            cryolo_mode=cryolo_mode,
            norm=norm_func,
            jitter=False,
            overlap=overlap_patches,
            name="valid",
            train_times=1,
            is_filament_training=filament_model
        )

        ############################################
        # Make a few callbacks
        ############################################

        all_callbacks = []

        early_stop = EarlyStopping(
            monitor="val_loss",
            min_delta=0.001,
            patience=early_stop_thresh,
            mode="min",
            verbose=1,
        )

        if parallel_model:
            checkpoint = MultiGPUModelCheckpoint(
                saved_weights_name,
                monitor="val_loss",
                verbose=1,
                save_best_only=True,
                save_weights_only=True,
                mode="min",
                period=1,
            )
            checkpoint.set_template_model(self.model)
        else:
            checkpoint = ExtendedModelCheckpoint(
                saved_weights_name,
                monitor="val_loss",
                verbose=1,
                save_best_only=True,
                save_weights_only=True,
                mode="min",
                period=1,
            )
        checkpoint.set_anchors(self.anchors)
        checkpoint.set_fine_tune(self.fine_tune, self.num_fine_tune_layers)
        checkpoint.set_version(cryolo.__version__)
        checkpoint.set_filamet_model(self.is_filament_model)

        reduceLROnPlateau = keras.callbacks.ReduceLROnPlateau(
            monitor="val_loss",
            factor=0.1,
            patience=int(early_stop_thresh * 0.6),
            verbose=1,
        )

        warmup_callback = WarmupCallback(
            warmup_bs=warmup_bs,
            num_warmup_epochs=warmup_epochs,
            len_train_img=len(train_imgs),
            len_valid_img=len(valid_imgs),
            batch_size=self.batch_size,
            num_patches=num_patches,
            train_times=train_times,
            valid_times=valid_times,
            early_stop_callback=early_stop,
            modelcheckpoint=checkpoint,
            reduceonnplateau=reduceLROnPlateau,
        )

        ##APPEND CALLBACKS

        all_callbacks.append(early_stop)
        all_callbacks.append(checkpoint)
        if not warm_restarts:
            all_callbacks.append(reduceLROnPlateau)

        tb_counter = (
            len(
                [
                    log
                    for log in os.listdir(os.path.expanduser(log_path))
                    if "yolo" in log
                ]
            )
            + 1
        )
        tensorboard = TensorBoard(
            log_dir=os.path.expanduser(log_path) + "yolo" + "_" + str(tb_counter),
            histogram_freq=0,
            write_graph=True,
            write_images=False,
        )
        all_callbacks.append(tensorboard)
        all_callbacks.append(warmup_callback)
        schedule = SGDRScheduler(
            min_lr=1e-7,
            max_lr=1e-4,
            steps_per_epoch=len(train_batch),
            lr_decay=0.9,
            cycle_length=5,
            mult_factor=1.5,
        )
        if warm_restarts:
            all_callbacks.append(schedule)

        num_workers = multiprocessing.cpu_count() // 2
        if num_cpus != -1:
            num_workers = num_cpus
        #self.use_multiprocessing=False
        num_workers = 1
        ############################################
        # Start the training process
        ############################################
        if parallel_model:
            parallel_model.fit_generator(
                generator=train_batch,
                steps_per_epoch=len(train_batch),
                epochs=nb_epoch,
                verbose=1,
                validation_data=valid_batch,
                validation_steps=len(valid_batch),
                callbacks=all_callbacks,
                workers=num_workers,
                max_queue_size=multiprocessing.cpu_count(),
                use_multiprocessing=self.do_multiprocessing,
            )
        else:
            self.model.fit_generator(
                generator=train_batch,
                steps_per_epoch=len(train_batch),
                epochs=nb_epoch,
                verbose=1,
                validation_data=valid_batch,
                validation_steps=len(valid_batch),
                callbacks=all_callbacks,
                workers=num_workers,
                max_queue_size=multiprocessing.cpu_count(),
                use_multiprocessing=self.do_multiprocessing,
            )
        # Rename to the user specified filename
        if os.path.exists(saved_weights_name):
            os.rename(saved_weights_name, saved_weights_filename + saved_weights_extension)



import unittest
import cryolo
import cryolo.CoordsIO
import cryolo.grouping3d
import os
import numpy as np


class MyTestCase(unittest.TestCase):
    def test_tracing_filament_3(self):

        from cryolo import utils
        from cryolo import grouping3d

        def generate_fil(length, offset=0):
            boxes = []

            for x in np.arange(offset, offset + length):
                box = utils.BoundBox(x=x, y=x, w=5, h=5, c=1)
                boxes.append(box)

            return utils.Filament(boxes)

        input_dict = {}
        input_dict[1] = [generate_fil(10, 0.5)]
        input_dict[2] = [generate_fil(10, 0.5)]

        filaments = grouping3d.do_tracing_filaments(
            input_dict,
            search_range=1,
            memory=2,
            min_edge_weight=0.3,
            window_size=2,
            resample_dist=1,
            merge_threshold=0.8,
        )

        self.assertEqual(1, len(filaments))

    def test_tracing_filament(self):
        """
        In this unit test I will generate the following filament pattern. Each row is a slice.
        - means filament position, x means no filament positions, - divided by x are seperate filaments.
        The result of the tracing should be 2 filaments.

        0123456789
        ----xx----
        ----------
        ----xx----
        xxxxxxxxx----xx----
        xxxxxxxxx----------
        xxxxxxxxx----xx----
        """
        from cryolo import utils
        from cryolo import grouping3d

        def generate_fil(length, offset=0):
            boxes = []
            for x in range(offset, offset + length):

                box = utils.BoundBox(x=x, y=0, w=5, h=5, c=1)
                boxes.append(box)

            return utils.Filament(boxes)

        input_dict = {}
        input_dict[1] = [generate_fil(4, 0), generate_fil(4, 6)]  # - - - - x x - - - -
        input_dict[2] = [generate_fil(10, 0)]
        input_dict[3] = [generate_fil(4, 0), generate_fil(4, 6)]
        input_dict[4] = [generate_fil(4, 9), generate_fil(4, 9 + 6)]
        input_dict[5] = [generate_fil(10, 9)]
        input_dict[6] = [generate_fil(4, 9), generate_fil(4, 9 + 6)]

        filaments = grouping3d.do_tracing_filaments(
            input_dict,
            search_range=1,
            memory=2,
            min_edge_weight=0.3,
            window_size=2,
            resample_dist=1,
            merge_threshold=0.8,
        )
        self.assertEqual(2, len(filaments))

    def test_tracing_filament_single_2(self):
        from cryolo import utils
        from cryolo import grouping3d

        def generate_fil(length, offset=0, c=1):
            boxes = []
            for x in range(offset, offset + length):
                box = utils.BoundBox(x=1.5, y=x, w=5, h=5, c=c)
                boxes.append(box)

            return utils.Filament(boxes)

        input_dict = {}
        input_dict[0] = [generate_fil(30, 5, 0.5)]
        input_dict[2] = [generate_fil(30, 0, 1)]

        filaments = grouping3d.do_tracing_filaments(
            input_dict,
            search_range=1,
            memory=0,
            min_edge_weight=0.3,
            window_size=1,
            resample_dist=1,
            merge_threshold=0.8,
        )
        print("Len:", len(filaments[0].boxes))
        for box in filaments[0].boxes:
            print(box.x, box.y, box.z)

        self.assertEqual(1, len(filaments))

    def test_tracing_filament_single(self):

        from cryolo import utils
        from cryolo import grouping3d

        def generate_fil(length, offset=0):
            boxes = []
            for x in range(offset, offset + length):
                box = utils.BoundBox(x=x, y=0, w=5, h=5, c=1)
                boxes.append(box)

            return utils.Filament(boxes)

        input_dict = {}
        input_dict[0] = [generate_fil(10, 0)]
        input_dict[2] = [generate_fil(5, 0)]

        filaments = grouping3d.do_tracing_filaments(
            input_dict,
            search_range=1,
            memory=0,
            min_edge_weight=0.3,
            window_size=2,
            resample_dist=1,
            merge_threshold=0.8,
        )
        for box in filaments[0].boxes:
            print(box.x, box.y, box.z)

        self.assertEqual(1, len(filaments))

    def test_filament_to_group3D_diag_1fil(self):
        import pandas as pd
        from cryolo import grouping3d

        path = os.path.join(
            os.path.dirname(__file__), "../resources/dataframe_diagonal_filament.csv"
        )
        df = pd.read_csv(path)
        fil3d = grouping3d.filaments_group_to_3D(df, window_size=90, box_distance=13)

        self.assertTrue(fil3d is not None, "Filament is none...")

    def test_tracing_filament_2(self):
        path = os.path.join(
            os.path.dirname(__file__), "../resources/filaments_untraced.cbox"
        )
        filaments = cryolo.CoordsIO.read_cbox_boxfile(path)

        input_dict = {}

        for fil in filaments:
            if int(fil.boxes[0].z) not in input_dict:
                input_dict[int(fil.boxes[0].z)] = [fil]
            else:
                input_dict[int(fil.boxes[0].z)].append(fil)
        from cryolo import grouping3d

        filaments_traced = grouping3d.do_tracing_filaments(
            input_dict,
            search_range=33.086 / 2,
            memory=2,
            min_edge_weight=0.0,
            window_size=33,
            min_length_filament=1,
            resample_dist=10,
        )

        self.assertEqual(3, len(filaments_traced))


if __name__ == "__main__":
    unittest.main()



import unittest
import os
from cryolo.evaluation.pathsboxdataprovider import PathsDataProvider
from cryolo.evaluation.pathsfilamentprovider import PathsFilamentProvider
from cryolo.evaluation.boxdataprovider import BoxDataProvider
import cryolo.CoordsIO as io
import glob


class MyTestCase(unittest.TestCase):
    def test_path_filament_provider_resampling(self):
        path_gt = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "../../resources/test_FolderDataProvider/01_gt_filament/*.cbox",
            )
        )
        files = glob.glob(path_gt)
        fdp5 = PathsFilamentProvider(
            paths=files,
            reading_method=io.read_cbox_boxfile,
            resampling_distance=1,
            group_method=BoxDataProvider.group_method_lastfolder,
        )
        fdp10 = PathsFilamentProvider(
            paths=files,
            reading_method=io.read_cbox_boxfile,
            resampling_distance=2,
            group_method=BoxDataProvider.group_method_lastfolder,
        )
        res5 = fdp5.get_boxes()
        res10 = fdp10.get_boxes()
        self.assertAlmostEquals(2, len(res5[0].data) / len(res10[0].data), delta=0.1)

    def test_path_filament_provider(self):
        path_gt = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "../../resources/test_FolderDataProvider/01_gt_filament/*.cbox",
            )
        )
        files = glob.glob(path_gt)
        fdp = PathsFilamentProvider(
            paths=files,
            reading_method=io.read_cbox_boxfile,
            group_method=BoxDataProvider.group_method_lastfolder,
        )
        res = fdp.get_boxes()
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0].data), 684 - 21 + 1)  # LAST_LINE - FIRST_LINE  + 1

    def test_get_correct_number(self):

        path_gt = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "../../resources/test_FolderDataProvider/00_gt/*.cbox",
            )
        )
        files = glob.glob(path_gt)
        fdp = PathsDataProvider(
            paths=files,
            reading_method=io.read_cbox_boxfile,
            group_method=BoxDataProvider.group_method_lastfolder,
        )
        res = fdp.get_boxes()

        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0].data), 3)


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import random
import time
from cryolo.augmentation.augmentation import Augmentation
from cryolo.augmentation.FlipAugmentation import FlipAugmentation
from cryolo.augmentation.CustomRotationAugmentation import CustomRotationAugmentation
from cryolo.augmentation.Rot90Augmentation import Rot90Augmentation
from cryolo.augmentation.AddConstantAugmentation import AddConstantAugmentation
from cryolo.augmentation.AdditiveGaussianNoiseAugmentation import (
    AdditiveGaussianNoiseAugmentation,
)
from cryolo.augmentation.AverageBlurAugmentation import AverageBlurAugmentation
from cryolo.augmentation.ContrastNormalizationAugmentation import (
    ContrastNormalizationAugmentation,
)
from cryolo.augmentation.DropoutAugmentation import DropoutAugmentation
from cryolo.augmentation.GaussBlurAugmentation import GaussBlurAugmentation


class AugTest(unittest.TestCase):
    def test_custom_rotation_augmentation_rot90_1x(self):
        img_width = 1024
        img_height = 1024
        boxwh = 100
        obj = {}
        box_center_x = 50
        box_center_y = 50

        obj["xmax"] = box_center_x + boxwh / 2
        obj["xmin"] = 0
        obj["ymax"] = box_center_y + boxwh / 2
        obj["ymin"] = 0
        obj["angle"] = np.pi / 2

        exp_obj = {}
        exp_obj["xmax"] = img_width - obj["ymin"]
        exp_obj["xmin"] = img_width - obj["ymax"]
        exp_obj["ymax"] = obj["xmax"]
        exp_obj["ymin"] = obj["xmin"]
        exp_obj["angle"] = obj["angle"] + np.pi / 2

        aug = Rot90Augmentation(1)

        aug_obj = aug.transform_coords([obj], image_dims=[img_width, img_height])[0]
        print(aug_obj)
        self.assertEqual(aug_obj["xmax"], exp_obj["xmax"], "Xmax is not the same")
        self.assertEqual(aug_obj["xmin"], exp_obj["xmin"], "Xmin is not the same")
        self.assertEqual(aug_obj["ymax"], exp_obj["ymax"], "ymax is not the same")
        self.assertEqual(aug_obj["ymin"], exp_obj["ymin"], "ymin is not the same")
        self.assertEqual(aug_obj["angle"], exp_obj["angle"], "angle is not the same")

    def test_custom_rotation_augmentation_rot90_4x(self):

        img_width = 1024
        img_height = 512
        boxwh = 100
        obj = {}
        box_center_x = 50
        box_center_y = 50

        obj["xmax"] = box_center_x + boxwh / 2
        obj["xmin"] = 0
        obj["ymax"] = box_center_y + boxwh / 2
        obj["ymin"] = 0
        obj["angle"] = np.pi / 2

        aug = Rot90Augmentation(4)

        aug_obj = aug.transform_coords([obj], image_dims=[img_width, img_height])[0]

        self.assertEqual(aug_obj["xmax"], obj["xmax"], "Xmax is not the same")
        self.assertEqual(aug_obj["xmin"], obj["xmin"], "Xmin is not the same")
        self.assertEqual(aug_obj["ymax"], obj["ymax"], "ymax is not the same")
        self.assertEqual(aug_obj["ymin"], obj["ymin"], "ymin is not the same")
        self.assertEqual(aug_obj["angle"], obj["angle"], "angle is not the same")

    def test_custom_rotation_augmentation_flip_vertical(self):

        img_width = 512
        img_height = 512
        boxwh = 100
        obj = {}
        obj["xmax"] = boxwh
        obj["xmin"] = 0
        obj["ymax"] = boxwh
        obj["ymin"] = 0
        obj["angle"] = np.pi / 2

        exp_obj = {}
        exp_obj["xmax"] = img_width - obj["xmin"]
        exp_obj["xmin"] = img_width - obj["xmax"]
        exp_obj["ymax"] = obj["ymax"]
        exp_obj["ymin"] = obj["ymin"]
        exp_obj["angle"] = np.pi - obj["angle"]
        aug = FlipAugmentation(FlipAugmentation.FLIP_VERTICAL)

        aug_obj = aug.transform_coords([obj], image_dims=[512, 512])[0]

        self.assertEqual(aug_obj["xmax"], exp_obj["xmax"], "Xmax is not the same")
        self.assertEqual(aug_obj["xmin"], exp_obj["xmin"], "Xmin is not the same")
        self.assertEqual(aug_obj["ymax"], exp_obj["ymax"], "ymax is not the same")
        self.assertEqual(aug_obj["ymin"], exp_obj["ymin"], "ymin is not the same")
        self.assertEqual(aug_obj["angle"], exp_obj["angle"], "angle is not the same")

    def test_custom_rotation_augmentation_rot90(self):

        rotation_center_x = 128
        rotation_center_y = 128
        boxwh = 10
        center_x = rotation_center_x + 10
        center_y = rotation_center_y + 0

        center_after_rot_x = rotation_center_x + 0
        center_after_rot_y = rotation_center_y - 10

        obj = {}
        obj["xmax"] = center_x + boxwh
        obj["xmin"] = center_x - boxwh
        obj["ymax"] = center_y + boxwh
        obj["ymin"] = center_y - boxwh
        obj["angle"] = np.pi / 2

        exp_obj = {}
        exp_obj["xmax"] = center_after_rot_x + boxwh
        exp_obj["xmin"] = center_after_rot_x - boxwh
        exp_obj["ymax"] = center_after_rot_y + boxwh
        exp_obj["ymin"] = center_after_rot_y - boxwh
        exp_obj["angle"] = np.pi / 2 - np.pi / 2
        cust_rot = CustomRotationAugmentation(np.deg2rad(90))
        rot_obj = cust_rot.transform_coords([obj], image_dims=[256, 256])[0]

        self.assertEqual(rot_obj["xmax"], exp_obj["xmax"], "Xmax is not the same")
        self.assertEqual(rot_obj["xmin"], exp_obj["xmin"], "Xmin is not the same")
        self.assertEqual(rot_obj["ymax"], exp_obj["ymax"], "ymax is not the same")
        self.assertEqual(rot_obj["ymin"], exp_obj["ymin"], "ymin is not the same")
        self.assertEqual(rot_obj["angle"], exp_obj["angle"], "angle is not the same")

    def test_custom_rotation_augmentation_rot360(self):
        rotation_center_x = 128
        rotation_center_y = 128
        boxwh = 10
        center_x = rotation_center_x + 10
        center_y = rotation_center_y + 0

        center_after_rot_x = rotation_center_x + 10
        center_after_rot_y = rotation_center_y + 0

        obj = {}
        obj["xmax"] = center_x + boxwh
        obj["xmin"] = center_x - boxwh
        obj["ymax"] = center_y + boxwh
        obj["ymin"] = center_y - boxwh
        obj["angle"] = np.pi / 4

        exp_obj = {}
        exp_obj["xmax"] = center_after_rot_x + boxwh
        exp_obj["xmin"] = center_after_rot_x - boxwh
        exp_obj["ymax"] = center_after_rot_y + boxwh
        exp_obj["ymin"] = center_after_rot_y - boxwh
        exp_obj["angle"] = np.pi / 4

        cust_rot = CustomRotationAugmentation(np.deg2rad(360))
        rot_obj = cust_rot.transform_coords([obj], image_dims=[256, 256])[0]

        self.assertEqual(rot_obj["xmax"], exp_obj["xmax"], "Xmax is not the same")
        self.assertEqual(rot_obj["xmin"], exp_obj["xmin"], "Xmin is not the same")
        self.assertEqual(rot_obj["ymax"], exp_obj["ymax"], "ymax is not the same")
        self.assertEqual(rot_obj["ymin"], exp_obj["ymin"], "ymin is not the same")
        self.assertEqual(rot_obj["angle"], exp_obj["angle"], "angle is not the same")

    def test_custom_rotation_augmentation_rot720(self):
        rotation_center_x = 128
        rotation_center_y = 128
        boxwh = 10
        center_x = rotation_center_x + 10
        center_y = rotation_center_y + 0

        center_after_rot_x = rotation_center_x + 10
        center_after_rot_y = rotation_center_y + 0

        obj = {}
        obj["xmax"] = center_x + boxwh
        obj["xmin"] = center_x - boxwh
        obj["ymax"] = center_y + boxwh
        obj["ymin"] = center_y - boxwh
        obj["angle"] = np.pi / 4

        exp_obj = {}
        exp_obj["xmax"] = center_after_rot_x + boxwh
        exp_obj["xmin"] = center_after_rot_x - boxwh
        exp_obj["ymax"] = center_after_rot_y + boxwh
        exp_obj["ymin"] = center_after_rot_y - boxwh
        exp_obj["angle"] = np.pi / 4
        cust_rot = CustomRotationAugmentation(np.deg2rad(720))
        rot_obj = cust_rot.transform_coords([obj], image_dims=[256, 256])[0]

        self.assertEqual(rot_obj["xmax"], exp_obj["xmax"], "Xmax is not the same")
        self.assertEqual(rot_obj["xmin"], exp_obj["xmin"], "Xmin is not the same")
        self.assertEqual(rot_obj["ymax"], exp_obj["ymax"], "ymax is not the same")
        self.assertEqual(rot_obj["ymin"], exp_obj["ymin"], "ymin is not the same")
        self.assertAlmostEqual(
            rot_obj["angle"], exp_obj["angle"], places=5, msg="angle is not the same"
        )

    def test_image_augmentation_time(self):
        np.random.seed(6)
        random.seed(10)
        img = np.random.random_sample(size=(4096, 4096)) * 255
        img = img.astype(np.uint8)
        img = img.astype(np.float32)
        start = time.time()
        aug = Augmentation(True)
        for i in range(10):
            aug.image_augmentation(img.copy())
            # aug.gauss_blur(img.copy())
            # print img.dtype
        end = time.time()

    def test_image_augmentation_float32(self):
        np.random.seed(6)
        random.seed(10)
        img = np.random.random_sample(size=(10, 10))
        img = img.astype(np.float32)
        aug = Augmentation()
        result = aug.image_augmentation(img.copy())
        is_float = np.issubdtype(result.dtype, np.float32)
        self.assertTrue(is_float, "Image augmentation failed. Type is not correct")

    def test_image_augmentation_uint8(self):
        np.random.seed(7)
        random.seed(10)
        img = np.random.randint(0, 255, size=(10, 10))
        img = img.astype(np.float32)
        aug = Augmentation(True)
        result = aug.image_augmentation(img.copy())
        is_int = np.issubdtype(result.dtype, np.uint8)
        self.assertTrue(is_int, "Image augmentation failed. Type is not correct")

    def test_additive_gauss_noise_typecheck_float32(self):
        img = np.random.random_sample(size=(3, 3))
        img = img.astype(np.float32)
        aug = AdditiveGaussianNoiseAugmentation(0.05)
        result = aug.transform_image(img.copy())
        is_float = np.issubdtype(result.dtype, np.float32)
        self.assertTrue(is_float, "ADD GAUSS NOISE FLOAT32 failed. Type is not correct")

    def test_avg_blur_float32(self):
        np.random.seed(7)
        img = np.random.random_sample(size=(3, 3))
        img = img.astype(np.float32)
        aug = AverageBlurAugmentation([3, 4])
        result = aug.transform_image(img.copy())
        t = np.allclose(np.mean(img), result[1, 1], atol=0.0001)
        self.assertTrue(t, "AVG BLUR FLOAT32 failed.")
        is_float = np.issubdtype(result.dtype, np.float32)
        self.assertTrue(is_float, "AVG BLUR FLOAT32 failed. Type is not correct")

    def test_avg_blur_uint8(self):
        np.random.seed(7)
        img = np.random.randint(0, 255, size=(3, 3))
        print(img)
        img = img.astype(np.uint8)
        aug = AverageBlurAugmentation([3, 4])
        result = aug.transform_image(img.copy())
        self.assertEqual(140, result[1, 1])
        is_int = np.issubdtype(result.dtype, np.uint8)
        self.assertTrue(is_int, "Avg blurring failed. Type is not correct")

    def test_gaussian_blur_float32(self):
        np.random.seed(7)
        img = np.random.random_sample(size=(3, 3))
        aug = GaussBlurAugmentation(sigma_range=[0, 3])
        result = aug.transform_image(img.copy())
        print(result)
        img_exp = [
            [0.3956204, 0.45140969, 0.4693832],
            [0.43998803, 0.44855732, 0.43916815],
            [0.43967383, 0.40600407, 0.37695657],
        ]
        t = np.allclose(img_exp, result, atol=0.00001)
        self.assertTrue(t, "Gaussian blurring failed")
        is_float = np.issubdtype(result.dtype, np.float32)

        self.assertTrue(is_float, "Gaussian blurring failed. Type is not correct")

    def test_gaussian_blur_uint8(self):
        np.random.seed(7)
        img = np.random.randint(0, 255, size=(3, 3))
        img = img.astype(np.float32)
        aug = GaussBlurAugmentation(sigma_range=[0, 3])
        result = aug.transform_image(img.copy())
        result = result.astype(np.uint8)
        print(result)
        img_exp = [[178, 144, 99], [175, 140, 125], [154, 125, 115]]
        t = np.array_equal(img_exp, result)
        self.assertTrue(t, "Gaussian blurring failed")

        is_int = np.issubdtype(result.dtype, np.integer)

        self.assertTrue(is_int, "Gaussian blurring failed. Type is not correct")

    def test_contrast_normalization_float32(self):
        np.random.seed(9)
        img = np.zeros((3, 3), dtype="float32")
        img[2, 2] = 255
        aug = ContrastNormalizationAugmentation([0.5, 2.0])
        print(img)
        result = aug.transform_image(img.copy())
        print(result)

        img_exp = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 131.46811]])

        t = np.allclose(img_exp, result, atol=0.0001)
        self.assertTrue(
            t, msg="Contrast normalization failed. Result is not as expected"
        )

        is_float = np.issubdtype(result.dtype, np.float32)

        self.assertTrue(is_float, "Contrast normalization failed. Type is not correct")

    def test_contrast_normalization_uint8(self):
        np.random.seed(9)
        img = np.zeros((3, 3), dtype="float32")
        img[2, 2] = 255
        aug = ContrastNormalizationAugmentation([0.5, 2.0])
        result = aug.transform_image(img.copy())
        print(result)
        result = result.astype(np.uint8)
        img_exp = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 131]])

        t = np.allclose(img_exp, result, atol=0.0001)

        self.assertTrue(
            t, msg="Contrast normalization failed. Result is not as expected"
        )

    def test_add_float32(self):
        np.random.seed(10)
        img = np.ones((3, 3), dtype="float32")
        img[0, 0] = 0
        img[2, 2] = 10
        aug = AddConstantAugmentation(0.1)
        result = aug.transform_image(img.copy())
        t = np.array_equal(img + 0.46959287, result)
        self.assertTrue(t, msg="Add failed. Result is not as expected")

        is_float = np.issubdtype(result.dtype, np.float32)

        self.assertTrue(is_float, "Add failed. Type is not correct")

    def test_add_uint8(self):
        np.random.seed(10)
        img = np.zeros((3, 3), dtype="float32")
        img[2, 2] = 255

        aug = AddConstantAugmentation(0.1)
        result = aug.transform_image(img.copy())
        result = np.clip(result, 0, 255)
        print(result)
        img_exp = img + 13.045981748296022
        img_exp[2, 2] = 255.0
        t = np.array_equal(img_exp, result)

        self.assertTrue(t, msg="Add failed. Result is not as expected")

    def test_multiply_float32(self):
        np.random.seed(10)
        img = np.ones((3, 3), dtype="float32")
        aug = Augmentation()
        result = aug.multiply(img.copy())
        img_exp = img * 1.27132064327
        t = np.array_equal(img_exp, result)
        self.assertTrue(
            t, msg="Float 32 multiplication failed. Result is not as expected"
        )

        is_float = np.issubdtype(result.dtype, np.float32)

        self.assertTrue(is_float, "Multiplication failed. Type is not correct")

    def test_multiply_uint8(self):
        np.random.seed(10)
        img = np.ones((3, 3), dtype="float32")
        img = img * 100
        img[2, 2] = 250
        aug = Augmentation(True)
        result = aug.multiply(img.copy())
        np.clip(result, 0, 255, out=result)
        result = result.astype(np.uint8, copy=False)

        img_exp = [[127, 127, 127], [127, 127, 127], [127, 127, 255]]

        t = np.array_equal(img_exp, result)
        self.assertTrue(t, msg="unit8 multiplication failed. Result is not as expected")

        is_int = np.issubdtype(result.dtype, np.uint8)

        self.assertTrue(is_int, "Multiplication failed. Type is not correct")

    def test_dropout_5x5_float32(self):
        np.random.seed(10)
        img = np.ones((6, 6), dtype="float32")
        img[:3, :] = 3
        aug = DropoutAugmentation(0.1)
        res = aug.transform_image(img.copy())
        num_zero_elements = np.sum(res == 2)
        self.assertEqual(3, num_zero_elements, "Dropout test failed.")

        is_float = np.issubdtype(res.dtype, np.float32)

        self.assertTrue(is_float, "Dropout failed. Type is not correct")

    def test_dropout_5x5_uint8(self):
        np.random.seed(10)
        img = np.ones((6, 6), dtype="float32")
        img[:3, :] = 3
        aug = DropoutAugmentation(0.1)
        res = aug.transform_image(img.copy())
        num_zero_elements = np.sum(res == 2)
        self.assertEqual(3, num_zero_elements, "Dropout test failed.")

    def test_dropout_10x10_uint8(self):
        np.random.seed(10)
        img = np.ones((10, 10), dtype="float32")
        img[:5, :] = 3
        aug = DropoutAugmentation(0.1)
        res = aug.transform_image(img.copy())
        num_zero_elements = np.sum(res == 2)
        self.assertEqual(8, num_zero_elements, "Dropout test failed.")



import unittest

import cryolo.utils


class MyTestCase(unittest.TestCase):
    def test_resample_filament_returns_fila(self):

        boxes = []
        boxes.append(cryolo.utils.BoundBox(x=1182, y=2116, w=30, h=30, c=1, classes=""))
        boxes.append(cryolo.utils.BoundBox(x=1248, y=1926, w=30, h=30, c=1, classes=""))
        fil = cryolo.utils.Filament(boxes)
        res_fil = cryolo.utils.resample_filament(fil, 10)
        self.assertTrue(res_fil.boxes is not None, "No boxes")

    def test_resample_filament_correct_length(self):

        boxes = []
        boxes.append(cryolo.utils.BoundBox(x=0, y=0, w=30, h=30, c=1, classes=""))
        boxes.append(cryolo.utils.BoundBox(x=100, y=0, w=30, h=30, c=1, classes=""))
        fil = cryolo.utils.Filament(boxes)
        res_fil = cryolo.utils.resample_filament(fil, 10)
        self.assertEqual(11, len(res_fil.boxes))

    def test_resample_filament_correct_length_1(self):

        boxes = []
        boxes.append(cryolo.utils.BoundBox(x=0, y=0, w=30, h=30, c=1, classes=""))
        boxes.append(cryolo.utils.BoundBox(x=10, y=0, w=30, h=30, c=1, classes=""))
        fil = cryolo.utils.Filament(boxes)
        res_fil = cryolo.utils.resample_filament(fil, 20)
        self.assertEqual(1, len(res_fil.boxes))

    def test_resample_filament_correct_length_2(self):

        boxes = []
        boxes.append(cryolo.utils.BoundBox(x=0, y=0, z=0, w=30, h=30, c=1, classes=""))
        boxes.append(cryolo.utils.BoundBox(x=1, y=0, z=0, w=30, h=30, c=1, classes=""))
        boxes.append(cryolo.utils.BoundBox(x=2, y=0, z=0, w=30, h=30, c=1, classes=""))
        boxes.append(cryolo.utils.BoundBox(x=3, y=0, z=0, w=30, h=30, c=1, classes=""))
        boxes.append(cryolo.utils.BoundBox(x=4, y=0, z=0, w=30, h=30, c=1, classes=""))
        boxes.append(cryolo.utils.BoundBox(x=5, y=0, z=1, w=30, h=30, c=1, classes=""))
        fil = cryolo.utils.Filament(boxes)
        res_fil = cryolo.utils.resample_filament(fil, 1)
        self.assertEqual(6, len(res_fil.boxes))

    def test_resample_filament_correct_maxx(self):

        boxes = []
        boxes.append(cryolo.utils.BoundBox(x=20, y=5, w=30, h=30, c=1, classes=""))
        boxes.append(cryolo.utils.BoundBox(x=85, y=35, w=30, h=30, c=1, classes=""))
        fil = cryolo.utils.Filament(boxes)
        res_fil = cryolo.utils.resample_filament(fil, 6)
        for b in res_fil.boxes:
            print(b.x, b.y)
        self.assertTrue(res_fil.boxes[-1].x <= 85)
        self.assertTrue(res_fil.boxes[0].x >= 20)
        self.assertTrue(res_fil.boxes[-1].y <= 35)
        self.assertTrue(res_fil.boxes[0].y >= 5)

    def test_find_corresponding_paths(self):
        path = "/u/v/w/a/b/c/d/file.mrc"
        list_of_paths = [
            "/u/v/w/a/b/c/e/file.mrc",
            "/u/v/w/a/b/c/d/file.mrc",
            "/u/v/w/a/f/c/d/file.mrc",
        ]
        pth = cryolo.utils.find_corresponding_path(list_of_paths, path)
        self.assertEqual(pth, path)

    def test_overlapping_boxes(self):
        def create_filament(flength, ystart, z, numbox, dy=1):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0, y=ystart + i * dy, z=z, w=10, h=10, depth=10, c=1, classes=[1]
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        fila = create_filament(flength=10, ystart=0, z=0, numbox=10, dy=10)
        filb = create_filament(flength=10, ystart=50, z=3, numbox=10, dy=10)

        fila = cryolo.utils.filament_to_array(fila)
        filb = cryolo.utils.filament_to_array(filb)
        r = cryolo.utils.overlapping_boxes_indicis(fila, filb, 0.3)

        self.assertEquals(len(r), 5)

    def test_non_maxima_suppression3D_average(self):
        # Test to try if averaging is working

        boxa = cryolo.utils.BoundBox(
            x=0, y=0, z=0, w=10, h=10, depth=10, c=1, classes=[1]
        )
        boxa.meta["num_boxes"] = 10
        boxb = cryolo.utils.BoundBox(
            x=1, y=1, z=1, w=10, h=10, depth=10, c=0.9, classes=[1]
        )
        boxb.meta["num_boxes"] = 10
        nms_thresh = 0.3
        obj_thresh = 0.3

        newboxes = cryolo.utils.non_maxima_suppress_fast_3d(
            [boxa, boxb], nms_thresh, obj_thresh
        )

        self.assertTrue(len(newboxes) == 1)
        self.assertTrue(newboxes[0].x == 0.5)
        self.assertTrue(newboxes[0].y == 0.5)
        self.assertTrue(newboxes[0].z == 0.5)

    def test_non_maxima_suppression3D_twoboxes(self):
        # Test to try if averaging is working

        boxa = cryolo.utils.BoundBox(
            x=0, y=0, z=0, w=10, h=10, depth=10, c=1, classes=[1]
        )
        boxa.meta["num_boxes"] = 10
        boxb = cryolo.utils.BoundBox(
            x=0, y=0, z=10, w=10, h=10, depth=10, c=0.9, classes=[1]
        )
        boxb.meta["num_boxes"] = 10
        nms_thresh = 0.3
        obj_thresh = 0.3

        newboxes = cryolo.utils.non_maxima_suppress_fast_3d(
            [boxa, boxb], nms_thresh, obj_thresh
        )

        self.assertTrue(len(newboxes) == 2)

    def test_merging_3D_filaments_1(self):
        def create_filament(flength, ystart, z, numbox, dy=1):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0, y=ystart + i * dy, z=z, w=10, h=10, depth=10, c=1, classes=[1]
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        fila = create_filament(flength=10, ystart=0, z=0, numbox=10, dy=10)
        filb = create_filament(flength=5, ystart=0, z=2, numbox=10, dy=10)
        nms_thresh = 0.3
        obj_thresh = 0.3

        newfils = cryolo.utils.merge_filaments_3d(
            [filb, fila], nms_thresh, obj_thresh, window_radius=1, box_distance=5
        )

        self.assertTrue(len(newfils) == 1, "Number of fils should be 1 but is" + str(len(newfils)))

    def test_get_closest_point(self):
        def create_filament(flength, ystart, z, numbox, dy=1, x=0, dz=0):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=x, y=ystart + i * dy, z=z +i*dz, w=10, h=10, depth=10, c=1, classes=[1]
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        fila = create_filament(flength=10, ystart=0, z=0, numbox=10, dy=10)
        filb = create_filament(flength=10, ystart=0, z=5, numbox=10, dy=10, dz=0.5)

        fildat_a = cryolo.utils.filament_to_array(fila)
        fildat_b = cryolo.utils.filament_to_array(filb)
        distance = cryolo.utils.get_closest_point_distance(fildat_a[:,:3],fildat_b[:,:3])
        print(distance)
        self.assertEqual(5,distance)



    def test_merging_3D_filaments_only_one(self):
        def create_filament(flength, ystart, z, numbox, dy=1):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0, y=ystart + i * dy, z=z, w=10, h=10, depth=10, c=1, classes=[1]
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        fila = create_filament(flength=10, ystart=0, z=0, numbox=10, dy=10)
        nms_thresh = 0.3
        obj_thresh = 0.3

        newfils = cryolo.utils.merge_filaments_3d(
            [fila], nms_thresh, obj_thresh, window_radius=1, box_distance=5
        )

        self.assertTrue(len(newfils) == 1, "Number of fils should be 1 but is" + str(len(newfils)))

    def test_merging_3D_filaments_many_for_timing(self):
        def create_filament(flength, ystart, z, numbox, dy=1):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0, y=ystart + i * dy, z=z, w=10, h=10, depth=10, c=1, classes=[1]
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)
        fils = []
        import numpy as np
        np.random.seed(100)
        nfilas = 500
        for i in range(nfilas):
            z = np.random.randint(0,300)
            fils.append(create_filament(flength=10, ystart=0, z=z, numbox=10, dy=10))

        nms_thresh = 0.3
        obj_thresh = 0.3

        newfils = cryolo.utils.merge_filaments_3d(
            fils, nms_thresh, obj_thresh, window_radius=1, box_distance=5
        )

        self.assertTrue(True)


    def test_merging_3D_filaments_partial(self):
        def create_filament(flength, ystart, z, numbox, dy=1):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0, y=ystart + i * dy, z=z, w=10, h=10, depth=10, c=1, classes=[1]
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        # Filaments are generated in that way. I expect 2 filaments with partial merging.
        # * * * * * * * * * *
        #           * * * * *
        #           * * * * * * * * * *
        fila = create_filament(flength=10, ystart=0, z=0, numbox=10, dy=10)
        filb = create_filament(flength=5, ystart=0, z=0, numbox=10, dy=10)
        filc = create_filament(flength=10, ystart=50, z=2, numbox=10, dy=10)
        nms_thresh = 0.3
        obj_thresh = 0

        newfils = cryolo.utils.merge_filaments_3d(
            [filb, fila, filc], nms_thresh, obj_thresh, window_radius=1, box_distance=10, partial=True
        )
        for f in newfils:
            print("Filament")
            for b in f.boxes:
                print(b.y)
        self.assertTrue(len(newfils) == 2, "Laenge " + str(len(newfils)))

    def test_merging_3D_filaments_partial2(self):
        def create_filament(flength, ystart, z, numbox, dy=1):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0, y=ystart + i * dy, z=z, w=10, h=10, depth=10, c=1, classes=[1]
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        # Filaments are generated in that way. I expect 2 filaments of length 10
        # * * * * * * * * * * L10
        #           * * * * * * * * * L9
        #                     * * * * * * * * * * L10
        fila = create_filament(flength=10, ystart=0, z=0, numbox=10, dy=10)
        filb = create_filament(flength=9, ystart=50, z=2, numbox=10, dy=10)
        filc = create_filament(flength=10, ystart=100, z=2, numbox=10, dy=10)

        nms_thresh = 0.3
        obj_thresh = 0

        newfils = cryolo.utils.merge_filaments_3d(
            [filb, fila, filc], nms_thresh, obj_thresh, window_radius=1, box_distance=10,
            partial=True
        )

        self.assertTrue(len(newfils) == 2, "Laenge " + str(len(newfils)))

    def test_merging_3D_filaments_partial3(self):
        def create_filament(flength, ystart, z, numbox, dy=1):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0, y=ystart + i * dy, z=z, w=10, h=10, depth=10, c=1, classes=[1]
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        # Filaments are generated in that way. I expect 2 filaments of length 10
        # * * * * * * * * * * L10
        #           * * * * * * * * * L9
        #                     * * * * * * * * * * L10
        fila = create_filament(flength=10, ystart=0, z=0, numbox=10, dy=10)
        filb = create_filament(flength=9, ystart=50, z=2, numbox=10, dy=10)
        filc = create_filament(flength=10, ystart=100, z=2, numbox=10, dy=10)
        fild = create_filament(flength=10, ystart=100, z=2, numbox=10, dy=10)
        file = create_filament(flength=10, ystart=100, z=2, numbox=10, dy=10)
        nms_thresh = 0.3
        obj_thresh = 0

        newfils = cryolo.utils.merge_filaments_3d(
            [filb, fila, filc, fild, file], nms_thresh, obj_thresh, window_radius=1, box_distance=10,
            partial=True
        )
        for f in newfils:
            print("Fialment")
            for b in f.boxes:
                print(b.y)
        self.assertTrue(len(newfils) == 2, "Laenge " + str(len(newfils)))

    def test_merging_3D_filaments_partial4(self):
        def create_filament(flength, ystart, z, numbox, dy=1):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0, y=ystart + i * dy, z=z, w=10, h=10, depth=10, c=1, classes=[1]
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        # Filaments are generated in that way. I expect 1 filament if m
        # * * * * * * * * * *           L10
        # * * * * * * * * * *           L10
        #         * * * * * * * * * *   L10
        filc = create_filament(flength=10, ystart=100, z=2, numbox=10, dy=10)
        fild = create_filament(flength=10, ystart=100, z=2, numbox=10, dy=10)
        file = create_filament(flength=10, ystart=140, z=2, numbox=10, dy=10)
        nms_thresh = 0.3
        obj_thresh = 0

        newfils = cryolo.utils.merge_filaments_3d(
            [filc, fild, file], nms_thresh, obj_thresh, window_radius=1, box_distance=10,
            partial=True
        )
        sum_length = 0 # should be 10 + 4
        for f in newfils:
            print("Fialment")
            sum_length += len(f.boxes)
            for b in f.boxes:
                print(b.y)
        self.assertTrue(len(newfils) == 2, f"Number of filament should be 2 but is {len(newfils)} ")
        self.assertTrue(sum_length == 14, f"Sum should be 12 but is {sum_length}")

    def test_merging_3D_filaments_partial5(self):
        def create_filament(flength, ystart, z, numbox, dy=1):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0, y=ystart + i * dy, z=z, w=10, h=10, depth=10, c=1, classes=[1]
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        # Filaments are generated in that way. I expect 1 filament if m
        # * * * * * * * * *           L9
        #         * * * * * * * * * *           L10
        #                 * * * * * * * * *    L9
        filc = create_filament(flength=9, ystart=100, z=2, numbox=10, dy=10)
        fild = create_filament(flength=10, ystart=140, z=2, numbox=10, dy=10)
        file = create_filament(flength=9, ystart=180, z=2, numbox=10, dy=10)
        nms_thresh = 0.3
        obj_thresh = 0

        newfils = cryolo.utils.merge_filaments_3d(
            [filc, fild, file], nms_thresh, obj_thresh, window_radius=1, box_distance=10,
            partial=True
        )
        sum_length = 0 # should be 10 + 4
        print("Test result:")
        for f in newfils:
            print(f"Fialment {len(f.boxes)}" )
            sum_length += len(f.boxes)
            for b in f.boxes:
                print(b.y)
        self.assertTrue(len(newfils) == 3, f"Number of filament should be 2 but is {len(newfils)} ")
        self.assertTrue(sum_length == (4+10+3), f"Sum should be {4+10+3} but is {sum_length}")

    def test_merging_3D_filaments_nonpartial5(self):
        def create_filament(flength, ystart, z, numbox, dy=1):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0, y=ystart + i * dy, z=z, w=10, h=10, depth=10, c=1, classes=[1]
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        # Filaments are generated in that way. I expect 1 filament if m
        # * * * * * * * * *           L9
        #         * * * * * * * * * *           L10
        #                 * * * * * * * * *    L9
        filc = create_filament(flength=9, ystart=100, z=2, numbox=10, dy=10)
        fild = create_filament(flength=10, ystart=140, z=2, numbox=10, dy=10)
        file = create_filament(flength=9, ystart=180, z=2, numbox=10, dy=10)
        nms_thresh = 0.3
        obj_thresh = 0

        newfils = cryolo.utils.merge_filaments_3d(
            [filc, fild, file], nms_thresh, obj_thresh, window_radius=1, box_distance=10,
            partial=False
        )
        sum_length = 0  # should be 10 + 4
        print("Test result:")
        for f in newfils:
            print(f"Fialment {len(f.boxes)}")
            sum_length += len(f.boxes)
            for b in f.boxes:
                print(b.y)
        self.assertTrue(len(newfils) == 1, f"Number of filament should be 1 but is {len(newfils)} ")
        self.assertTrue(sum_length == 17,
                        f"Sum should be {17} but is {sum_length}")

    def test_merging_3D_filaments_nonpartial4(self):
        def create_filament(flength, ystart, z, numbox, dy=1):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0, y=ystart + i * dy, z=z, w=10, h=10, depth=10, c=1, classes=[1]
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        fila = create_filament(flength=10, ystart=100, z=2, numbox=10, dy=10)
        filc = create_filament(flength=10, ystart=100, z=2, numbox=10, dy=10)
        fild = create_filament(flength=10, ystart=100, z=2, numbox=10, dy=10)
        file = create_filament(flength=10, ystart=100, z=2, numbox=10, dy=10)
        nms_thresh = 0.3
        obj_thresh = 0

        newfils = cryolo.utils.merge_filaments_3d(
            [fila, filc, fild, file], nms_thresh, obj_thresh, window_radius=1,
            box_distance=10,
            partial=False
        )
        for f in newfils:
            print("Fialment")
            for b in f.boxes:
                print(b.y)
        self.assertTrue(len(newfils) == 1, "Laenge " + str(len(newfils)))

    def test_get_consecutive_subarrays(self):
        # 0 1 2 3 4 5 6 7 8 9 10 11
        # * * o o o * * o o o * *
        import numpy as np
        overlap = np.array([2,3,4,7,8,9])
        splt = cryolo.utils.get_consecutive_subarrays(overlap)
        np.testing.assert_array_equal(splt[0], np.array([2, 3, 4]))
        np.testing.assert_array_equal(splt[1], np.array([7, 8, 9]))

    def test_merging_3D_filaments_length_Correct(self):
        def create_filament(flength, ystart, z, numbox, dy=1, bsize=10):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0,
                    y=ystart + i * dy,
                    z=z,
                    w=bsize,
                    h=bsize,
                    depth=bsize,
                    c=1,
                    classes=[1],
                )
                box.meta["num_boxes"] = numbox
                print(i, box.x, box.y)
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        dy = 2
        fila = create_filament(flength=10, ystart=0, z=0, numbox=10, dy=dy, bsize=3)
        filb = create_filament(
            flength=10, ystart=5 * dy, z=1, numbox=10, dy=dy, bsize=3
        )

        merged = cryolo.utils.merge_filaments_3d(
            [fila, filb],
            iou_threshold=0.3,
            merge_threshold=0.4,
            window_radius=dy,
            box_distance=dy,
        )

        self.assertEqual(1, len(merged))
        self.assertEqual(15, len(merged[0].boxes))

    def test_resampling_filament_noeffect(self):
        def create_filament(flength, ystart, z, numbox, dy=1, bsize=10):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0,
                    y=ystart + i * dy,
                    z=z,
                    w=bsize,
                    h=bsize,
                    depth=bsize,
                    c=1,
                    classes=[1],
                )
                box.meta["num_boxes"] = numbox
                print(i, box.x, box.y)
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        dy = 2
        fila = create_filament(flength=10, ystart=0, z=0, numbox=10, dy=dy, bsize=3)
        rfila = cryolo.utils.resample_filament(fila, dy)
        self.assertEqual(len(fila.boxes), len(rfila.boxes))

    def test_merging_3D_filaments_2(self):
        def create_filament(flength, ystart, z, numbox, dy=1, bsize=10):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0,
                    y=ystart + i * dy,
                    z=z,
                    w=bsize,
                    h=bsize,
                    depth=bsize,
                    c=1,
                    classes=[1],
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        dy = 2
        fila = create_filament(flength=10, ystart=0, z=0, numbox=10, dy=dy, bsize=3)
        filb = create_filament(
            flength=10, ystart=5 * dy, z=1, numbox=10, dy=dy, bsize=3
        )

        merged = cryolo.utils.merge_filaments_3d(
            [fila, filb],
            iou_threshold=0.3,
            merge_threshold=0.4,
            window_radius=3 * dy,
            box_distance=dy,
        )

        self.assertEqual(1, len(merged))

    def test_iou_boxes_shifted(self):
        import numpy as np

        box1 = cryolo.utils.BoundBox(x=100, y=100, w=50, h=50, c=1, classes=[1])
        box2 = cryolo.utils.BoundBox(x=100 + 25, y=100, w=50, h=50, c=1, classes=[1])
        iou = cryolo.utils.bbox_iou(box1, box2)
        self.assertEqual(iou, 1.0 / 3.0)

    def test_iou_boxes_same(self):
        box1 = cryolo.utils.BoundBox(x=100, y=100, w=50, h=50, c=1, classes=[1])
        box2 = cryolo.utils.BoundBox(x=100, y=100, w=50, h=50, c=1, classes=[1])
        iou = cryolo.utils.bbox_iou(box1, box2)
        self.assertEqual(iou, 1.0)

    def test_merging_3D_filaments_3(self):
        def create_filament(flength, ystart, z, numbox, dy=1, bsize=10):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0,
                    y=ystart + i * dy,
                    z=z,
                    w=bsize,
                    h=bsize,
                    depth=bsize,
                    c=1,
                    classes=[1],
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        dy = 2
        fila = create_filament(flength=10, ystart=0, z=0, numbox=10, dy=dy, bsize=5)
        filb = create_filament(
            flength=10, ystart=5 * dy, z=1, numbox=10, dy=dy, bsize=5
        )
        filc = create_filament(flength=1, ystart=0, z=2, numbox=10, dy=dy, bsize=5)

        merged = cryolo.utils.merge_filaments_3d(
            [fila, filb, filc],
            iou_threshold=0.3,
            merge_threshold=0.4,
            window_radius=dy,
            box_distance=dy,
        )

        self.assertEqual(1, len(merged))
        self.assertEqual(15, len(merged[0].boxes))

    def test_merging_3D_filaments_posweight(self):
        def printfil(fil):
            for box in fil.boxes:
                print(box.x, box.y, box.z, box.c)

        def create_filament(flength, ystart, z, numbox, dy=1, bsize=10):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0,
                    y=ystart + i * dy,
                    z=z,
                    w=bsize,
                    h=bsize,
                    depth=bsize,
                    c=1,
                    classes=[1],
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        dy = 5
        fila = create_filament(flength=10, ystart=0, z=0, numbox=2, dy=dy, bsize=6)
        filb = create_filament(flength=10, ystart=0, z=3, numbox=8, dy=dy, bsize=6)

        merged = cryolo.utils.merge_filaments_3d(
            [fila, filb],
            iou_threshold=0.3,
            merge_threshold=0.4,
            window_radius=3 * dy,
            box_distance=dy,
        )

        self.assertEqual(10, merged[0].boxes[0].meta["num_boxes"])
        self.assertEqual(2.4, merged[0].boxes[0].z)

    def test_merging_3D_filaments_partialmerge(self):
        def printfil(fil):
            for box in fil.boxes:
                print(box.x, box.y, box.z)

        def create_filament(flength, ystart, z, numbox, dy=1, bsize=10):
            fboxes = []
            for i in range(flength):
                box = cryolo.utils.BoundBox(
                    x=0,
                    y=ystart + i * dy,
                    z=z,
                    w=bsize,
                    h=bsize,
                    depth=bsize,
                    c=1,
                    classes=[1],
                )
                box.meta["num_boxes"] = numbox
                fboxes.append(box)
            return cryolo.utils.Filament(fboxes)

        dy = 2
        fila = create_filament(flength=10, ystart=0, z=0, numbox=10, dy=dy, bsize=3)
        filb = create_filament(
            flength=20, ystart=5 * dy, z=1, numbox=10, dy=dy, bsize=3
        )

        merged = cryolo.utils.merge_filaments_3d(
            [fila, filb],
            iou_threshold=0.3,
            merge_threshold=0,
            window_radius=dy,
            box_distance=dy,
            partial=True,
        )

        import numpy as np

        print("Found:", len(merged), len(merged[0].boxes), len(merged[1].boxes))
        printfil(merged[0])
        print(" ")
        printfil(merged[1])
        sorted = np.sort([len(merged[0].boxes), len(merged[1].boxes)])
        self.assertEqual(20, sorted[1])
        self.assertEqual(5, sorted[0])


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod

import numpy as np


class AugmentationMethod(ABC):

    @abstractmethod
    def transform_image(self, image : np.array) -> np.array:
        pass

    @abstractmethod
    def transform_coords(self, object_coords : list, image_dims : list) -> list:
        pass


from cryolo.augmentation.AugmentationMethod import AugmentationMethod

import numpy as np

class AddConstantAugmentation(AugmentationMethod):

    def __init__(self, scale : float):
        """
        :param scale: Scale for random constant. The random constant
        will be between 0 and scale*6*std(image)
        """
        self.scale = scale

    def transform_image(self, image : np.array) -> np.array:
        """
        Adds a random constant to the image
        :param image: Input image
        :return: Modified image
        """
        width = 2 * 3 * np.std(image)
        width_rand = self.scale * width
        rand_constant = (np.random.rand() * width_rand) - width_rand / 2
        np.add(image, rand_constant, out=image)
        return image

    def transform_coords(self, object_coords : list, image_dims : list) -> list:
        return object_coords